/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CAOSDB_CONNECTION_H
#define CAOSDB_CONNECTION_H
/**
 * @file caosdb/connection.h
 * @author Timm Fitschen
 * @date 2021-05-18
 * @brief Configuration and setup of the connection.
 */
#include <iosfwd>                                // for ostream
#include <map>                                   // for map
#include <memory>                                // for shared_ptr, unique_ptr
#include <string>                                // for string, basic_string
#include "caosdb/authentication.h"               // for Authenticator
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionSe...
#include "caosdb/info.h"                         // for VersionInfo
#include "caosdb/info/v1alpha1/main.grpc.pb.h"   // for GeneralInfoService:...
#include "caosdb/transaction.h"                  // for Transaction
#include "grpcpp/channel.h"                      // for Channel
#include "grpcpp/security/credentials.h"         // for ChannelCredentials

namespace caosdb::connection {
using boost::filesystem::path;
using caosdb::authentication::Authenticator;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::transaction::Transaction;
using grpc::ChannelCredentials;

class CertificateProvider {
public:
  [[nodiscard]] auto virtual GetCertificatePem() const -> std::string = 0;
  virtual ~CertificateProvider() = default;
};

class PemFileCertificateProvider : public CertificateProvider {
private:
  std::string certificate_provider;

public:
  explicit PemFileCertificateProvider(const path &path);
  [[nodiscard]] auto GetCertificatePem() const -> std::string override;
};

class PemCertificateProvider : public CertificateProvider {
private:
  std::string certificate_provider;

public:
  explicit PemCertificateProvider(const std::string &certificate_provider);
  [[nodiscard]] auto GetCertificatePem() const -> std::string override;
};

/**
 * @brief Configuration of the CaosDB connection.
 */
class ConnectionConfiguration {
private:
  std::string host;
  int port;

public:
  ConnectionConfiguration(const std::string &host, int port);
  virtual ~ConnectionConfiguration() = default;
  friend auto operator<<(std::ostream &out,
                         const ConnectionConfiguration &configuration)
    -> std::ostream &;

  [[nodiscard]] auto virtual ToString() const -> std::string = 0;
  [[nodiscard]] auto GetHost() const -> std::string;
  [[nodiscard]] auto GetPort() const -> int;
  [[nodiscard]] auto virtual GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> = 0;
};

class InsecureConnectionConfiguration : public ConnectionConfiguration {
private:
  std::shared_ptr<ChannelCredentials> credentials;

public:
  InsecureConnectionConfiguration(const std::string &host, int port);
  [[nodiscard]] auto GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> override;
  [[nodiscard]] auto ToString() const -> std::string override;
};

class TlsConnectionConfiguration : public ConnectionConfiguration {
private:
  std::shared_ptr<ChannelCredentials> credentials;
  std::string certificate_provider;

public:
  TlsConnectionConfiguration(const std::string &host, int port);
  TlsConnectionConfiguration(const std::string &host, int port,
                             const Authenticator &authenticator);
  TlsConnectionConfiguration(const std::string &host, int port,
                             const CertificateProvider &certificate_provider);
  TlsConnectionConfiguration(const std::string &host, int port,
                             const CertificateProvider &certificate_provider,
                             const Authenticator &authenticator);
  [[nodiscard]] auto GetChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> override;
  [[nodiscard]] auto ToString() const -> std::string override;
};

/**
 * @brief A reusable connection to a CaosDBServer.
 */
class Connection {
  std::shared_ptr<grpc::Channel> channel;
  std::unique_ptr<GeneralInfoService::Stub> general_info_service;
  std::shared_ptr<EntityTransactionService::Stub> entity_transaction_service;

public:
  explicit Connection(const ConnectionConfiguration &configuration);
  friend auto operator<<(std::ostream &out, const Connection &connection)
    -> std::ostream &;
  [[nodiscard]] auto GetVersionInfo() const -> std::unique_ptr<VersionInfo>;
  [[nodiscard]] auto CreateTransaction() const -> std::unique_ptr<Transaction>;
};

/**
 * Lazily creates and caches reusable connection instances. Singleton.
 *
 * This class delegates the configuration of new connections to the global
 * ConfigurationManager.
 *
 * A reset of the ConfigurationManager also resets the ConnectionManager.
 *
 * @brief Lazily creates and caches reusable connection instances.
 */
class ConnectionManager {
private:
  mutable std::map<std::string, std::shared_ptr<Connection>> connections;
  mutable std::string default_connection_name;
  inline ConnectionManager(){};

  auto mHasConnection(const std::string &name) const -> bool;

  auto mGetConnection(const std::string &name) const
    -> const std::shared_ptr<Connection> &;

  auto mGetDefaultConnection() const -> const std::shared_ptr<Connection> &;

  inline auto mReset() -> void {
    connections.clear();
    default_connection_name = std::string();
  }

public:
  static ConnectionManager &GetInstance() {
    static ConnectionManager instance;
    return instance;
  };

  inline static auto HasConnection(const std::string &name) -> bool {
    return ConnectionManager::GetInstance().mHasConnection(name);
  };

  inline static auto GetConnection(const std::string &name)
    -> const std::shared_ptr<Connection> & {
    return ConnectionManager::GetInstance().mGetConnection(name);
  };

  inline static auto GetDefaultConnection()
    -> const std::shared_ptr<Connection> & {
    return ConnectionManager::GetInstance().mGetDefaultConnection();
  };

  inline static auto Reset() -> void {
    return ConnectionManager::GetInstance().mReset();
  };

  ConnectionManager(ConnectionManager const &) = delete;
  void operator=(ConnectionManager const &) = delete;
};

} // namespace caosdb::connection
#endif
